# Multiple Python environments in one Renku session

## Manage Python environments
There are many ways to create different virtual environments in one session, for example, `virtualenv`, `pyenv`, etc. In this example, we utilize `conda` to manage two Python environments:

- 1st environment: `base`
    - Python version: **3.9**
    - Dependencies: `environment.yml`
    - This environment is the default environment. It comes with the base image `renku/renkulab-py:3.9-0.10.1`. You do not need to create it yourself.
- 2nd environment: `py3.6-sklearn-env`
    - Python version: **3.6**
    - Dependencies: `py3.6-sklearn-env.yml`
    - This environment is created by Lines 27-34 of `Dockerfile`.
    - It contains Python 3.6 with three additional libraries:
      - `ipykernel`: required if you want to have a Jupyter kernel with this environment
      - `scikit-learn`, `matplotlib`: for demo purpose

## Use Jupyter with different Python
After installing `ipykernel` and registering the 2nd environment to Jupyter (see Line 32 of `Dockerfile`), you can create Jupyter notebooks with different kernels:

<img src="img/create_notebook.png" width="250">

Also, you can change the kernel after creating the notebook with the drop-down menu in the top right corner of the notebook:

<img src="img/change_kernel.png" width="400">


## Create Renku workflow with different Python

### EXAMPLE 1: Same script, different python
Script `ex1.py` outputs the Python path and the time of execution to a file named with the Python version, `ex1_output_{version}.txt`.

```bash
renku run --name=base_run -- ${CONDA_PREFIX}/bin/python ex1.py
# create file ex1_output_3.9.6.txt
```

```bash
conda activate py3.6-sklearn-env
renku run --name=py36_run -- ${CONDA_PREFIX}/bin/python ex1.py
# create file ex1_output_3.6.15.txt
conda deactivate
```

### EXAMPLE 2: Two-stage two-Python workflow

In this example, we create a two-stage workflow.

**Stage 1** uses `base` envrionment (Python 3.9). `ex2_stage1.py` reads the input file `ex2_input.txt`, removes the prefix 'Renku_' of each sentence, and outputs the results to the file `ex2_stage1_output.txt`.

```bash
renku run --name=remove_prefix -- ${CONDA_PREFIX}/bin/python ex2_stage1.py ex2_input.txt
# create file ex2_stage1_output.txt
```

**Stage 2** uses `py3.6-sklearn-env` environment (Python 3.6). `ex2_stage2.py` reads the output file of stage 1, vectorizes each input sentence, calculate the pair-wise cosine similarity and outputs the similarity matrix as an image `ex2_stage2_output.png`.

```bash
conda activate py3.6-sklearn-env
renku run --name=similarity_viz -- ${CONDA_PREFIX}/bin/python ex2_stage2.py ex2_stage1_output.txt
# create file ex2_stage2_output.png
conda deactivate
```

**Note:** to make the example more representative:
- **Stage 1** can only be run in `base`, as it requires a feature (`str.removeprefix(prefix)`) first introduced in Python 3.9
- **Stage 2** can only be run in `py3.6-sklearn-env`, as it requires `scikit-learn` and `matplotlib` that do not exist in `base`

**Workflow visualization:**

```
  ┌─────────────┐       ┌─────────────┐                 
  │ex2_stage1.py│       │ex2_input.txt│                 
  └─────────────┘       └─────────────┘                 
           *                 **                         
            **             **                           
              **         **                             
╔══════════════════════════════════════╗                
║/opt/conda/bin/python ex2_stage1.py...║                
╚══════════════════════════════════════╝                
                   *                                    
                   *                                    
                   *                                    
         ┌─────────────────────┐       ┌─────────────┐  
         │ex2_stage1_output.txt│       │ex2_stage2.py│  
         └─────────────────────┘       └─────────────┘  
                      **                   **           
                        **               **             
                          **           **               
                ╔═════════════════════════════════╗     
                ║/opt/conda/envs/py3.6-sklearn-...║     
                ╚═════════════════════════════════╝     
                                 *                      
                                 *                      
                                 *                      
                      ┌─────────────────────┐           
                      │ex2_stage2_output.png│           
                      └─────────────────────┘
```